package businesscalendar;

import java.util.Date;
import java.util.ArrayList;
import java.util.Arrays;
import java.text.SimpleDateFormat;
import java.text.ParsePosition;
import java.util.GregorianCalendar;

import java.text.SimpleDateFormat;


/**
 * Title:        BusinessCalendar
 * Description:  Business calendar using summation to
 *               quickly calculate arbitrary business time between different dates.
 *
 * @author Efraim Berkovich
 * @version 1.0
 */

public class BusinessCalendar
{
  public static long MSECS_PER_DAY                        = 1000 * 60 * 60 * 24;
  public static long MSECS_PER_HOUR                       = 1000 * 60 * 60;
  public static long MSECS_PER_MINUTE                     = 1000 * 60;

  public static long MSECS_PER_WORKING_DAY                = MSECS_PER_HOUR * 8;

  /** to hold the list of dates, bus. start , bus. end */
  private Date[] date;

  /** to hold the summation matching the dates */
  private long[] sum;

  /** to speedup the check for whether the date[] date is start of period or end */
  private boolean[] isStart;

  /** start and end dates of calendar */
  private Date start, end;

  private Date[] daysOff;


  /**
   * Create a simple business calendar. Rem: Weekends are
   * automatically non-business days. Business time is 9-5.
   *
   * @param start Date which is the first day of the calendar
   * @param daysOff Date[] of holidays (i.e. non-business days)
   * @param end Date which is the last date of the calendar
   *
   * @exception IllegalArgumentException if parameters don't make sense
   */
  public BusinessCalendar( Date start, Date[] daysOff, Date end )
  {
    this.start  = start;
    this.end = end;

    // verify parameters
    if( end.before( start ) )
      throw new IllegalArgumentException( "End date must be after start date" );
    if( daysOff == null )
      throw new IllegalArgumentException( "daysOff cannot be null" );

    // ensure daysOff is in ascending order
    this.daysOff = daysOff;
    Arrays.sort( this.daysOff );

    ArrayList cal = new ArrayList();
    ArrayList summation = new ArrayList();

    long day = start.getTime();
    long lastSum = 0;

    cal.add( new Date( day ) );
    summation.add( new Long( lastSum ) );

    Date d = new Date( day );

    while( d.before( end ) )
    {
      if( ! isDayOff( d ) )
      {
        // add for working day
        Date xx = makeDate( new Date( day ), "09:00" );  // starts at 9
        long z = xx.getTime();

        cal.add( new Date( z ) );
        summation.add( new Long( lastSum ) );

        z = z + ( 8 * MSECS_PER_HOUR );  // working day 8 hours
        lastSum = lastSum + ( 8 * MSECS_PER_HOUR );
        cal.add( new Date( z ) );
        summation.add( new Long( lastSum ) );
      }
      day = day + MSECS_PER_DAY;
      d.setTime( day );
    }
    cal.add( new Date( end.getTime() ) );
    summation.add( new Long( lastSum ) );

    date = (Date[])cal.toArray( new Date[ cal.size() ] );
    sum = new long[ date.length ];
    isStart = new boolean[ date.length ];

    isStart[ 0 ] = (! isDayOff( start ));
    sum[ 0 ] = 0;

    for( int i = 1; i < date.length; i++ )
    {
      sum[ i ] = ((Long)summation.get( i )).longValue();
      isStart[ i ] = (sum[ i ] > sum[ i - 1 ]);
    }

  }


  /**
   * make a date from Date (for day part) and String time (for time part)
   */
  private static Date makeDate( Date d, String time )
  {
     // Format the current time.
     SimpleDateFormat f1
         = new SimpleDateFormat ("yyyy-MM-dd ");
     String dateString = f1.format( d );
     dateString += time;

     SimpleDateFormat f2
        = new SimpleDateFormat( "yyyy-MM-dd HH:mm" );

     ParsePosition pos = new ParsePosition(0);
     return f2.parse(dateString, pos);
   }

  /**
   * Check if holiday or weekend
   * @param d Date to check
   * @return true if non-business day
   */
  private boolean isDayOff( Date d )
  {
    int n = Arrays.binarySearch( daysOff, d );
    if( n < 0 )    // not found
    {
      // check if a weekend
      GregorianCalendar c = new GregorianCalendar();
      c.setTime( d );
      int w = c.get( GregorianCalendar.DAY_OF_WEEK );
      if ( ( w == 7 ) || ( w == 1 ) )
        return true; // is weekend
      else
        return false; // is regular day
    }
    else // is holiday
    {
      return true;
    }
  }



  /**
   * find the F(x)
   * @param Date d
   */
  private long findSum( Date d )
  {
    int d1 = Arrays.binarySearch( date, d );
    if( d1 < 0 )
      d1 = Math.abs( d1 ) - 1;

    if( isStart[ d1 ] )
    {
      long offset = d.getTime() - date[ d1 ].getTime();
      return sum[ d1 ] + offset;
    }
    else
    {
      return sum[ d1 ];
    }
  }


  /**
   * find the F-1(x)
   * @param Date d
   */
  private Date findInverse( long n )
  {
    if( n < 0 )
      throw new IndexOutOfBoundsException( "Result precedes start of calendar." );

    int d1 = Arrays.binarySearch( sum, n );
    if( d1 < 0 )
      d1 = Math.abs( d1 ) - 1;

    if( d1 >= date.length )
      throw new IndexOutOfBoundsException( "Result exceeds end of calendar." );

    long val = 0;
    if( isStart[ d1 ] )
    {
      long offset = n - sum[ d1 ];
      val = val + offset;
    }

    val = val + date[ d1 ].getTime();
    return new Date( val );
  }




  /**
   * Check if a particular date is during business time
   * @param d Date to check
   * @return True if the date is during business time
   *
   * @exception IllegalArgumentException if d is not within the calendar
   */
  public boolean isBusinessDate( Date d )
  {
    if( d.before( this.start ) )
      throw new IllegalArgumentException( "Dates must be after calendar start." );
    if( d.after( this.end ) )
      throw new IllegalArgumentException( "Dates must be before calendar end." );

    int pos = Arrays.binarySearch( date, d );
    pos = Math.abs( pos );

    return this.isStart[ pos ];
  }


  /**
   * Returns the amount of business time between two dates.
   * @param date1 Date to compare
   * @param date2 Date to compare
   *
   * @return number of business time (in millis) between the two dates
   *
   * @exception IllegalArgumentException if dates are outside the calendar
   */
  public long diff( Date date1, Date date2 )
  {
    if(   ( date1.before( this.start ) )
       || ( date2.before( this.start ) )
      )
      throw new IllegalArgumentException( "Dates must be after calendar start." );
    if(   ( date1.after( this.end ) )
       || ( date2.after( this.end ) )
      )
      throw new IllegalArgumentException( "Dates must be before calendar end." );

    long d1 = findSum( date1 );
    long d2 = findSum( date2 );

    return d2 - d1;
  }


  /**
   * Add an amount of business time to a calendar date.
   * @param d Date at which to start
   * @param businessTime long of millis to add (or negative to subrract)
   *
   * @return Date of the business time businessTime from d date
   *
   * @exception IllegalArgumentException if date is outside of calendar
   * @exception IndexOutOfBoundsException if result is outside of calendar
   */
  public Date add( Date d, long businessTime )
  {
    if( d.before( this.start ) )
      throw new IllegalArgumentException( "from date must be after calendar start." );
    if( d.after( this.end ) )
      throw new IllegalArgumentException( "from date must be before calendar end." );

    long f1 = findSum( d );
    f1 = f1 + businessTime;

    return findInverse( f1 );
  }





  public static void main( String[] argv )
  {
    try
    {
      SimpleDateFormat f = new SimpleDateFormat( "M-d-yyyy" );
      Date start = f.parse( "01-01-2002" );
      Date end = f.parse( "01-01-2003" );
      Date[] daysOff = new Date[] { f.parse( "01-01-2002" )
                                  , f.parse( "02-14-2002" )
                                  , f.parse( "03-15-2002" ) };

      BusinessCalendar cal = new BusinessCalendar( start, daysOff, end );


      Date d1 = f.parse( "1-10-2002 10:00" );
      Date d2 = f.parse( "1-15-2002 15:30" );
      long numBusinessTime = 236 * BusinessCalendar.MSECS_PER_WORKING_DAY;

      System.out.println( "Is business date " + d1.toString() + "?   "
                          + cal.isBusinessDate( d1 ) );

      System.out.println( "Diff between " + d1.toString() + " and "
                          + d2.toString() + " is " + cal.diff( d1, d2 ) + " msecs"
                          + " or " + (cal.diff( d1, d2 )/cal.MSECS_PER_WORKING_DAY) + " working days." );

      System.out.println( "Add " + numBusinessTime + " business msecs to " + d1.toString()
                          + " is " + cal.add( d1, numBusinessTime ) );

    }
    catch( Exception e )
    {
      e.printStackTrace();
    }
  }

}