package businesscalendar;

import java.util.Date;
import java.util.Arrays;
import java.text.SimpleDateFormat;
import java.util.GregorianCalendar;

/**
 * Title:        DayBusinessCalendar
 * Description:  Simple business calendar using summation to
 *               quickly calculate business days between different dates.
 *
 * @author Efraim Berkovich
 * @version 1.0
 */

public class DayBusinessCalendar
{
  private static long MSECS_PER_DAY = 1000 * 60 * 60 * 24;

  // this will hold the summation results
  private int[] calendar;
  private Date[] inverse;

  private Date start;
  private Date end;
  private Date[] daysOff;

  private float density;

  /**
   * Create a simple business calendar. Rem: Weekends are
   * automatically non-business days.
   *
   * @param start Date which is the first day of the calendar
   * @param daysOff Date[] of holidays (i.e. non-business days)
   * @param end Date which is the last date of the calendar
   *
   * @exception IllegalArgumentException if parameters don't make sense
   */
  public DayBusinessCalendar( Date start, Date[] daysOff, Date end )
  {
    this.start  = start;
    this.end = end;

    // verify parameters
    if( end.before( start ) )
      throw new IllegalArgumentException( "End date must be after start date" );
    if( daysOff == null )
      throw new IllegalArgumentException( "daysOff cannot be null" );

    // ensure daysOff is in ascending order
    this.daysOff = daysOff;
    Arrays.sort( this.daysOff );

    long n = end.getTime() - start.getTime();
    calendar = new int[ (int)(n / MSECS_PER_DAY) ];
    if( calendar.length < 1 )
      throw new IllegalArgumentException( "Calendar must be at least one day long" );

    inverse = new Date[ calendar.length ];

    long day = start.getTime();
    Date d = new Date( day );

    calendar[ 0 ] = findCount( d );
    inverse[ calendar[ 0 ] ] = new Date( day );

    day = day + MSECS_PER_DAY;
    d.setTime( day );

    for( int i = 1; i < calendar.length; i++ )
    {
      calendar[ i ] = calendar[ i - 1 ] + findCount( d );
      if( inverse[ calendar[ i ] ] == null )    // keep only the smallest date
        inverse[ calendar[ i ] ] = new Date( day );

      day = day + MSECS_PER_DAY;
      d.setTime( day );
    }

    // find the density of business days in the calendar
    density = (float)calendar[ calendar.length - 1 ]/(float)calendar.length;
  }


  /**
   * Used to determine whether a day is a business day or not. This is
   * in effect, the count function returning 0 or 1.
   * @param d Date to check
   * @return 1 if a business day, 0 otherwise
   */
  private int findCount( Date d )
  {
    int n = Arrays.binarySearch( daysOff, d );
    if( n < 0 )    // not found
    {
      // check if a weekend
      GregorianCalendar c = new GregorianCalendar();
      c.setTime( d );
      int w = c.get( GregorianCalendar.DAY_OF_WEEK );
      if ( ( w == 7 ) || ( w == 1 ) )
        return 0; // is weekend
      else
        return 1; // is ok
    }
    else // is holiday
    {
      return 0;
    }
  }


  /**
   * Check if a particular day is a business day by
   * checking if the integraion increases or not.
   * @param d Date to check
   * @return True if the date is a business day
   *
   * @exception IllegalArgumentException if d is not within the calendar
   */
  public boolean isBusinessDate( Date d )
  {
    if( d.before( this.start ) )
      throw new IllegalArgumentException( "Dates must be after calendar start." );
    if( d.after( this.end ) )
      throw new IllegalArgumentException( "Dates must be before calendar end." );

    long offset = d.getTime() - start.getTime();
    offset = offset / MSECS_PER_DAY;
    if( offset == 0 )
      return ( calendar[ 0 ] == 1 );

    int n = (int)offset;
    return ( calendar[ n ] != calendar[ n - 1 ] );
  }


  /**
   * Returns the number of business days between two dates.
   * @param date1 Date to compare
   * @param date2 Date to compare
   *
   * @exception IllegalArgumentException if dates are outside the calendar
   */
  public int diff( Date date1, Date date2 )
  {
    if(   ( date1.before( this.start ) )
       || ( date2.before( this.start ) )
      )
      throw new IllegalArgumentException( "Dates must be after calendar start." );
    if(   ( date1.after( this.end ) )
       || ( date2.after( this.end ) )
      )
      throw new IllegalArgumentException( "Dates must be before calendar end." );

    long offset1 = date1.getTime() - start.getTime();
    offset1 = offset1 / MSECS_PER_DAY;

    long offset2 = date2.getTime() - start.getTime();
    offset2 = offset2 / MSECS_PER_DAY;

    int f1 = calendar[ (int)offset1 ];
    int f2 = calendar[ (int)offset2 ];

    return f2 - f1;
  }


  /**
   * Add a number of business days to a calendar date by searching consecutively.
   * @param d Date at which to start
   * @param numBusinessDays int of days to add (or negative to subrract)
   * @return Date of the business day numBusinessDays from d date
   * @exception IllegalArgumentException if date is outside of calendar
   * @exception IndexOutOfBoundsException if result is outside of calendar
   */
  public Date add3( Date d, int numBusinessDays )
  {
    if( d.before( this.start ) )
      throw new IllegalArgumentException( "from date must be after calendar start." );
    if( d.after( this.end ) )
      throw new IllegalArgumentException( "from date must be before calendar end." );

    long offset = d.getTime() - start.getTime();
    int d1 = (int)(offset/ MSECS_PER_DAY);
    int f1 = calendar[ d1 ];

    int d2 = (int)( numBusinessDays / density );
    d2 = d2 + d1;
    if( d2 >= calendar.length )
      throw new IndexOutOfBoundsException( "Result exceeds end of calendar." );
    if( d2 < 0 )
      throw new IndexOutOfBoundsException( "Result precedes start of calendar." );
    int f2 = calendar[ d2 ];

    int direction = 0;
    if(  ( f2 - f1 ) < numBusinessDays )
      direction = 1;
    else
      if(  ( f2 - f1 ) > numBusinessDays )
        direction = -1;

    while( true )
    {
      d2 = d2 + direction;
      if( d2 >= calendar.length )
        throw new IndexOutOfBoundsException( "Result exceeds end of calendar." );
      if( d2 < 0 )
        throw new IndexOutOfBoundsException( "Result precedes start of calendar." );

      f2 = calendar[ d2 ];

      if( ( f2 - f1 ) == numBusinessDays )
        break;
    }

    // since non-unique, move to first
    if( d2 == 0 )
      return start;

    while( calendar[ d2 - 1 ] == calendar[ d2] )
    {
      d2--;
      if( d2 == 0 ) return start;
    }
    return new Date( (long)( d2 * MSECS_PER_DAY ) + start.getTime() );
  }



  /**
   * Add a number of business days to a calendar date (using inverse found by binary search).
   * @param d Date at which to start
   * @param numBusinessDays int of days to add (or negative to subrract)
   * @return Date of the business day numBusinessDays from d date
   * @exception IllegalArgumentException if date is outside of calendar
   * @exception IndexOutOfBoundsException if result is outside of calendar
   */
  public Date add2( Date d, int numBusinessDays )
  {
    if( d.before( this.start ) )
      throw new IllegalArgumentException( "from date must be after calendar start." );
    if( d.after( this.end ) )
      throw new IllegalArgumentException( "from date must be before calendar end." );

    long offset = d.getTime() - start.getTime();
    int d1 = (int)(offset/ MSECS_PER_DAY);
    int f1 = calendar[ d1 ];

    int d2 = Arrays.binarySearch( calendar, f1 + numBusinessDays );
    if( d2 >= calendar.length )
      throw new IndexOutOfBoundsException( "Result exceeds end of calendar." );
    if( d2 < 0 )
      throw new IndexOutOfBoundsException( "Result precedes start of calendar." );

    // since non-unique, move to first
    if( d2 == 0 )
      return start;

    while( calendar[ d2 - 1 ] == calendar[ d2] )
    {
      d2--;
      if( d2 == 0 ) return start;
    }
    return new Date( (long)( d2 * MSECS_PER_DAY ) + start.getTime() );
  }


  /**
   * Add a number of business days to a calendar date (using inverse directly).
   * @param d Date at which to start
   * @param numBusinessDays int of days to add (or negative to subrract)
   * @return Date of the business day numBusinessDays from d date
   * @exception IllegalArgumentException if date is outside of calendar
   * @exception IndexOutOfBoundsException if result is outside of calendar
   */
  public Date add( Date d, int numBusinessDays )
  {
    if( d.before( this.start ) )
      throw new IllegalArgumentException( "from date must be after calendar start." );
    if( d.after( this.end ) )
      throw new IllegalArgumentException( "from date must be before calendar end." );

    long offset = d.getTime() - start.getTime();
    int d1 = (int)(offset/ MSECS_PER_DAY);
    int f1 = calendar[ d1 ];

    int d2 = f1 + numBusinessDays;
    if( d2 >= inverse.length )
      throw new IndexOutOfBoundsException( "Result exceeds end of calendar." );
    if( d2 < 0 )
      throw new IndexOutOfBoundsException( "Result precedes start of calendar." );

    Date r = inverse[ d2 ];
    if( r == null )
      throw new IndexOutOfBoundsException( "Result exceeds end of calendar." );

    return r;
  }



  public static void main( String[] argv )
  {
    try
    {
      SimpleDateFormat f = new SimpleDateFormat( "M-d-yyyy" );
      Date start = f.parse( "01-01-2002" );
      Date end = f.parse( "01-01-2003" );
      Date[] daysOff = new Date[] {
                                    f.parse( "01-01-2002" )
                                  , f.parse( "02-14-2002" )
                                  , f.parse( "03-15-2002" )
                                  , f.parse( "07-04-2002" )
                                  , f.parse( "09-02-2002" )
                                  , f.parse( "12-10-2002" )
                                    };

      DayBusinessCalendar cal = new DayBusinessCalendar( start, daysOff, end );


      Date d1 = f.parse( "01-10-2002" );
      Date d2 = f.parse( "12-13-2002" );
      int numBusinessDays = 238;

     System.out.println( "Is business date " + d1.toString() + "?   "
                          + cal.isBusinessDate( d1 ) );

     System.out.println( "Diff between " + d1.toString() + " and "
                          + d2.toString() + " is " + cal.diff( d1, d2 ) );

      System.out.println( "Add (using inverse) " + numBusinessDays + " business days to " + d1.toString()
                          + " is " + cal.add( d1, numBusinessDays ) );

      System.out.println( "Add (using binary search) " + numBusinessDays + " business days to " + d1.toString()
                          + " is " + cal.add2( d1, numBusinessDays ) );

      System.out.println( "Add (using consecutive search) " + numBusinessDays + " business days to " + d1.toString()
                          + " is " + cal.add3( d1, numBusinessDays ) );
    }
    catch( Exception e )
    {
      e.printStackTrace();
    }
  }

}