/////////////////////////////////////////////////////////////////////////////
// Example source code accompanying "Contract Programming and RTTI"
//
// Copyright 2002, Timothy M. Shead
//
// Permission is granted to use this code without restriction as long as this
// copyright appears in all source files.

#ifndef HOMEGROWN_RTTI_H
#define HOMEGROWN_RTTI_H

#include <cassert>

// This is a (very basic) framework for implementing a Contract Programming model
// without Standard C++ RTTI ... interfaces are derived from "Unknown", each concrete object
// derives from the interfaces it supports, and implements QueryInterface.

class Unknown
{
public:
	// InterfaceIDs are unique integers used to identify the interface
	// being queried for in a  call to QueryInterface() ... you will have
	// to maintain your own repository of unique InterfaceIDs as new
	// interfaces are defined over time.
	typedef unsigned long InterfaceID;
	
	// The unique ID for this interface ... you should hide it with an identical member in derived interfaces.
	static const InterfaceID iid;

	// Implement this in each derived concrete object to return supported interfaces.
	// You will have to maintain each implementation as your objects' interfaces change over time.
	// Returns true iff the requested InterfaceID is supported.  The "Interface" pointer will contain
	// the requested interface if supported, will be unchanged if not.
	virtual bool QueryInterface(const InterfaceID IID, void** Interface) = 0;
			
protected:
	virtual ~Unknown() {}
};

// Convenience template that allows you to call QueryInterface() as if it were dynamic_cast - 
// returns the requested interface if supported, or NULL
template<class InterfaceType>
InterfaceType* QueryInterface(Unknown* const Source)
{
	if(!Source)
		return 0;
		
	InterfaceType* result = 0;
	Source->QueryInterface(InterfaceType::iid, reinterpret_cast<void**>(&result));
	
	return result;
}

// Convenience template for implementing Unknown::QueryInterface() in concrete classes -
// use it to test the requested InterfaceID against supported interfaces
template<class InterfaceType>
bool SupportInterface(InterfaceType* const Server, const Unknown::InterfaceID IID, void** Interface)
{
	// Sanity checks ...
	assert(Interface);
	
	if(InterfaceType::iid != IID)
		return false;
		
	*Interface = static_cast<void*>(Server);
	return true;
}

#endif // HOMEGROWN_RTTI_H


