#ifndef HOMEGROWN_RTTI_SHAPES_H
#define HOMEGROWN_RTTI_SHAPES_H

#include <iostream>

// Second version of the sample drawing application, where the
// Contract Programming model has been applied, but "homegrown" RTTI
// is used in the form of QueryInterface(), instead of using dynamic_cast

// Note that the homegrown RTTI framework is complex enough to
// merit its own header:
#include "homegrown_rtti.h"

// Plus, all of the InterfaceIDs have to be maintained in a separate compilation unit -
// in this case, they're in "homegrown_rtti_ids.cpp", which will have to be linked into
// all implementations and all clients

// Abstract interface for shapes that can be drawn
class Drawable :
	public virtual Unknown
{
public:
	// This must be provided for every interface
	static const InterfaceID iid;

	// Implement this in derived classes so the shape can draw itself
	virtual void Draw() = 0;
	
protected:
	virtual ~Drawable() {}
	Drawable& operator=(const Drawable& rhs) {}
};

// Abstract interface for shapes that can be serialized to a stream
class Serializable :
	public virtual Unknown
{
public:
	// This must be provided for every interface
	static const InterfaceID iid;

	// Saves the shape to a stream
	virtual void Save(std::ostream& stream) = 0;
	
protected:
	virtual ~Serializable() {}
	Serializable& operator=(const Serializable& rhs) {}
};

// Abstract interface for shapes that can be interactively selected by the user
class Selectable :
	public virtual Unknown
{
public:	
	// This must be provided for every interface
	static const InterfaceID iid;

	// Selects this shape
	virtual void Select() = 0;
	// Deselects this shape
	virtual void Deselect() = 0;
	// Return selection state of this shape
	virtual bool IsSelected() = 0;
	
protected:
	virtual ~Selectable() {}
	Selectable& operator=(const Selectable& rhs) {}
};
	
// Abstract interface for shapes that can be positioned within the drawing
class Movable :
	public virtual Unknown
{
public:
	// This must be provided for every interface
	static const InterfaceID iid;

	// Moves the shape to a new position
	virtual void Move(const double x, const double y) = 0;

protected:
	virtual ~Movable() {}
	Movable& operator=(const Movable& rhs) {}
};

// Abstract interface for shapes that can be deleted by the user
class Deletable :
	public virtual Unknown
{
public:
	// This must be provided for every interface
	static const InterfaceID iid;

	virtual ~Deletable() {}
	Deletable& operator=(const Deletable& rhs) {}
};

// Container class that manages a drawing "canvas" full of shapes
class Canvas
{
public:
	~Canvas()
	{
		for(Shapes::iterator shape = m_Shapes.begin(); shape != m_Shapes.end(); ++shape)
			DeleteShape(*shape);
	}
	
	void Draw()
	{
		for(Shapes::iterator shape = m_Shapes.begin(); shape != m_Shapes.end(); ++shape)
			{
				Drawable* const drawable = QueryInterface<Drawable>(*shape);
				if(drawable)
					drawable->Draw();
			}
	}

	void Save(std::ostream& Stream)
	{
		for(Shapes::iterator shape = m_Shapes.begin(); shape != m_Shapes.end(); ++shape)
			{
				Serializable* const serializable = QueryInterface<Serializable>(*shape);
				if(serializable)
					serializable->Save(Stream);
			}
	}

	void AddShape(Unknown* const Shape)
	{
		assert(Shape);
		m_Shapes.push_back(Shape);
	}

private:
	void DeleteShape(Unknown* const Shape)
	{
		Deletable* const deletable = QueryInterface<Deletable>(Shape);
		if(deletable)
			delete deletable;
	}

	typedef std::vector<Unknown*> Shapes;
	Shapes m_Shapes;
};

// Convenience function for deselecting shapes
inline void DeselectShape(Unknown* const Shape)
{
	Selectable* const selectable = QueryInterface<Selectable>(Shape);
	if(selectable)
		selectable->Deselect();
}

#endif // HOMEGROWN_RTTI_SHAPES_H


