#ifndef SHAPES_H
#define SHAPES_H

#include <cassert>
#include <iostream>

// Our initial sample drawing application, where all drawing objects derive
// from the single abstract "Shape" class, which is cluttered with
// a laundry-list of unrelated & optional methods

// Abstract base class that lumps-together all possible shape capabilities 
class Shape
{
public:
	virtual ~Shape() {}

	// Returns true iff this shape can be drawn
	virtual bool IsDrawable() = 0;
	// Implement this in derived classes so the shape can draw itself
	virtual void Draw() = 0;
	
	// Returns true iff this shape can be serialized
	virtual bool IsSerializable() = 0;
	// Saves the shape to a stream
	virtual void Save(std::ostream& Stream) = 0;

	// Returns true iff this shape can be selected/deselected by the user
	virtual bool IsSelectable() = 0;
	// Selects the shape
	virtual void Select() = 0;
	// Deselects the shape
	virtual void Deselect() = 0;
	// Returns true iff this shape is selected
	virtual bool IsSelected() = 0;
			
	// Returns true iff this shape can be moved by the user
	virtual bool IsMovable() = 0;
	// Moves the shape to a new position
	virtual void Move(const double X, const double Y) = 0;
	
	// Returns true iff this object can be deleted
	virtual bool IsDeletable() = 0;
	
	/* Many more methods here */
	
protected:
	Shape& operator=(const Shape& RHS) {}
};

// Container class that manages a drawing "canvas" full of shapes
class Canvas
{
public:
	~Canvas()
	{
		for(Shapes::iterator shape = m_Shapes.begin(); shape != m_Shapes.end(); ++shape)
			DeleteShape(*shape);
	}
	
	void Draw()
	{
		for(Shapes::iterator shape = m_Shapes.begin(); shape != m_Shapes.end(); ++shape)
			if((*shape)->IsDrawable())
				(*shape)->Draw();
	}

	void Save(std::ostream& Stream)
	{
		for(Shapes::iterator shape = m_Shapes.begin(); shape != m_Shapes.end(); ++shape)
			if((*shape)->IsSerializable())
				(*shape)->Save(Stream);
	}

	void AddShape(Shape* const S)
	{
		assert(S);
		m_Shapes.push_back(S);
	}

private:
	void DeleteShape(Shape* const S)
	{
		assert(S);
		if(S->IsDeletable())
			delete S;
	}

	typedef std::vector<Shape*> Shapes;
	Shapes m_Shapes;
};

// Convenience function for deselecting a shape
inline void DeselectShape(Shape* const S)
{
	if(S->IsSelectable())
		S->Deselect();
}

#endif // SHAPES_H


