#ifndef SHAPES_H
#define SHAPES_H

#include <iostream>

// The final sample drawing application, in which related sets of functionality
// have been separated into abstract interfaces that can be tested for at runtime
// with dynamic_cast<>()

class Unknown
{
protected:
	virtual ~Unknown() {}
	Unknown& operator=(const Unknown& rhs) {}
};

class Drawable :
	public virtual Unknown
{
public:
	// Implement this in derived classes so the shape can draw itself
	virtual void Draw() = 0;
	
protected:
	virtual ~Drawable() {}
	Drawable& operator=(const Drawable& rhs) {}
};

class Serializable :
	public virtual Unknown
{
public:
	// Saves the shape to a stream
	virtual void Save(std::ostream& stream) = 0;
	
protected:
	virtual ~Serializable() {}
	Serializable& operator=(const Serializable& rhs) {}
};
	
class Selectable :
	public virtual Unknown
{
public:	
	// Selects this shape
	virtual void Select() = 0;
	// Deselects this shape
	virtual void Deselect() = 0;
	// Returns the selection state for this shape
	virtual bool IsSelected() = 0;
	
protected:
	virtual ~Selectable() {}
	Selectable& operator=(const Selectable& rhs) {}
};

class Movable :
	public virtual Unknown
{
public:
	// Moves the shape to a new position
	virtual void Move(const double x, const double y) = 0;

protected:
	virtual ~Movable() {}
	Movable& operator=(const Movable& rhs) {}
};

class Deletable :
	public virtual Unknown
{
public:
	virtual ~Deletable() {}
	Deletable& operator=(const Deletable& rhs) {}
};

// Container class that manages a drawing "canvas" full of shapes
class Canvas
{
public:
	~Canvas()
	{
		for(Shapes::iterator shape = m_Shapes.begin(); shape != m_Shapes.end(); ++shape)
			DeleteShape(*shape);
	}
	
	void Draw()
	{
		for(Shapes::iterator shape = m_Shapes.begin(); shape != m_Shapes.end(); ++shape)
			{
				Drawable* const drawable = dynamic_cast<Drawable*>(*shape);
				if(drawable)
					drawable->Draw();
			}
	}

	void Save(std::ostream& Stream)
	{
		for(Shapes::iterator shape = m_Shapes.begin(); shape != m_Shapes.end(); ++shape)
			{
				Serializable* const serializable = dynamic_cast<Serializable*>(*shape);
				if(serializable)
					serializable->Save(Stream);
			}
	}

	void AddShape(Unknown* const Shape)
	{
		assert(Shape);
		m_Shapes.push_back(Shape);
	}

private:
	void DeleteShape(Unknown* const Shape)
	{
		Deletable* const deletable = dynamic_cast<Deletable*>(Shape);
		if(deletable)
			delete deletable;
	}

	typedef std::vector<Unknown*> Shapes;
	Shapes m_Shapes;
};

// Convenience function for deselecting shapes
inline void DeselectShape(Unknown* const Shape)
{
	Selectable* const selectable = dynamic_cast<Selectable*>(Shape);
	if(selectable)
		selectable->Deselect();
}

#endif // SHAPES_H


